#!/usr/bin/env python

"""
Module to read segy files.  No writing at this time.

To test: './segy.py -v'


     Copyright (C) 2005  Kurt Schwehr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Id: segy.py,v 1.17 2005/05/30 16:51:57 schwehr Exp $

Eventually, each segy version (e.g. xstar/edgetech, and other
instrument tweaks) will be loaded on the fly.  So if a field is
somewhere else, it will be easy to tweak a new config in just a few
minutes

file:///sw/share/doc/python24/html/api/importing.html
"""
import mmap
import os
import struct	# Convert two and from binary formats

import unittest


#traceHdrSize = 240 # bytes

# These are the format characters for integers with the struct module for binary conversions
# Force big endian (default on osx, right?)
structIntCodes = { 1: '>b', 2: '>h', 4: '>i', 8: '>q' }


#
# FIX: load segy_rev1.py
#



class Segy:
    """
    Base class that contains the SEGY header and all the traces

    http://seg.org/publications/tech-stand/
    Then look at segy rev 0 and rev 1
    """

    def __init__(self,filename=None,drivername='segy_rev1'):
        """
        Open up a segy file for reading using mmap for speed.

        filename - name of the segy file to read
        drivername - what format of segy does this file need.
        	FIX: Eventually a value of None should allow you to have this
                code identify the version.
        """
        # Dynamically load the driver for the version of segy
        # FIX: why am I forcing the loading of the source.  Can this be done better?
        import imp
        driverFile = open(drivername+'.py','r')
        self.drv = imp.load_module(drivername,driverFile,drivername+'.py',('.py','r',imp.PY_SOURCE))
        
        file = open(filename,"r+")
        self.size = os.path.getsize(filename)
        self.data = mmap.mmap(file.fileno(),self.size,access=mmap.ACCESS_READ)
        file.close() # No longer need the file handle
        self.header=SegyHdr(self.data,self.size,drv=self.drv)
        return

    def getHeader(self):
        return self.header

    def getTraceHdr(self,traceNumber=1):
        if 1==traceNumber:
            return SegyTraceHdr(segyHdr=self.header,mmapData=self.header.data,startOffset=3600,drv=self.drv)
        extTextSize = 0 # FIX!!!
        traceStart = 3600 + extTextSize + (traceNumber-1)*self.getTraceSize()
        return SegyTraceHdr(segyHdr=self.header,mmapData=self.header.data,startOffset=traceStart, drv=self.drv)

    def getTraceDataSize(self,traceNumber=1):
        """ FIX: handle variable length traces"""
        format = self.header.getBinEntry('SampleFormat')
        f = self.drv.dataFormatStruct[format]
        size = struct.calcsize(f)
        samples = self.header.getBinEntry('SamplesPerTrace')
        datasize = size * samples
        return datasize

    def getTraceSize(self, traceNumber=1):
        """ returns the number of bytes for the particular trace

        fix: cope with variable length traces
        """
        return 240 + self.getTraceDataSize()

    def getNumberOfTraces(self):
        """ FIX: deal with variable length traces """
        extTextSize = 0 # FIX
        tracesSize = self.size - 3600 - extTextSize
        numTraces = tracesSize / self.getTraceSize()
        if 0 != (tracesSize % self.getTraceSize()):
            print "WARNING: seems you have variable length traces that will not be handled!!!"
            assert(False)
        return numTraces
    def getTraceData(self,traceNumber=1):
        """ return a list of samples

        FIX: cope with variable length traces
        """
        extTextSize = 0 # Calc this!
        traceStart = 3600 + extTextSize + (traceNumber-1)*self.getTraceSize()
        print "traceStart = ", traceStart, traceNumber, self.getTraceSize(), (traceNumber-1)*self.getTraceSize()

        startData = 240 + traceStart
        endData = startData+self.getTraceDataSize()

        format = self.header.getBinEntry('SampleFormat')
        formatStr = self.drv.dataFormatStruct[format]
        parseStr = formatStr[0] + formatStr[1] * self.header.getBinEntry('SamplesPerTrace')

        rawData = self.data[startData:endData]

        traceData = struct.unpack(parseStr,rawData)

        return traceData

    def writeTraceToFile(self,filename='default.dat',traceNumber=1):
        """ Write out pairs of samplenumber and value to a file
        """
        data = self.getTraceData(traceNumber)
        o = open(filename,"w")
        for i in range(0,len(data)):
            o.write(str(i)+" "+str(data[i])+"\n")
        o.close()
        
######################################################################
class SegyHdr:
    """  class for reading segy seismic file header

    http://seg.org/publications/tech-stand/
    Then look at segy rev 0 and rev 1
    """
    def __init__(self,mmapData=None,size=None, filename=None,drv=None):
        """
        Open up a segy file for reading using mmap for speed.
        """
        assert(drv)
        self.drv = drv
        if filename and (not (mmapData and size)):
            file = open(filename,"r+")
            self.size = os.path.getsize(filename)
            self.data = mmap.mmap(file.fileno(),self.size,access=mmap.ACCESS_READ)
            file.close() # No longer need the file handle
            return
        assert(mmapData and size)
        self.data=mmapData
        self.size=size
        return

    # FIX: move this to Segy class
    def hasR1TextFileHeader(self):
        """Turn True if it has a R1 header with C1..C40 sections as ASCII"""
        if self.data[0:2] == 'C1' or self.data[0:3] == 'C 1':
            return True
        return False

    # FIX: move this to Segy class
    def getTextHeaderLine(self,lineNumber):
        """ Return the string for the line number """
        line = self.data[(lineNumber-1)*80:(lineNumber)*80]
        nul = line.find("\0")
        if -1 == nul: return line
        return line[0:nul]
    
    # FIX: move this to Segy class  or make a text header class
    def getTextHeaderField(self, fieldName):
        """
        BROKEN

        e.g.:
        
        C2  LINE:5t6         AREA:Goleta
        """
        print "FIX: broken"
        entry = textFileHeaderEntries[fieldName]
        line = self.getTextHeaderLine(entry[0])
        if len(line) < entry[1]:
            print "Field "+fieldName+" not present"
            return None
        print "FIX: figure out the end of the entry"
        
        return line[entry[1]:]
        
    def getBinEntry(self,fieldName):
        """ Retrieve the value from one of the binary header entries by name """
        assert(0<len(fieldName))

        entry  = self.drv.binaryHeaderEntries[fieldName]
        start  = entry[0] - 1
        end    = entry[1] - 1 
        length = entry[1]-entry[0] + 1
        if length > 8:
            print "get bin header... too large.  skipping ",fieldName
            return None
        val = struct.unpack(structIntCodes[length],self.data[start:end+1])[0]

        return val

    def getTextTranslateValue(self,key,val):
        """ Lookup in the driver what the translation for a value should be

        e.g. 'SampleFormat' of 1 will return:
        4-byte IBM floating point

        """
        table = self.drv.fileHeaderTables[key]
        return table[val]

    def printBinaryHeaders(self):
        """ Dump all header into to stdout"""
        for key in self.drv.binaryHeaderEntries.keys():
            val = self.getBinEntry(key)
            print key, " = ",val

        #key = 'SampleFormat'; print key,'  =>  ',dataFormats[self.getBinEntry(key)]
        key = 'SampleFormat'; print key,'  =>  ',self.getTextTranslateValue(key,self.getBinEntry(key))
        
######################################################################
# Trace parsing
######################################################################
class SegyTraceHdr:
    """
    Handle one trace in a segy file.
    """
    def __init__(self, segyHdr, startOffset,mmapData=None, drv=None):
        """
        Load up a trace from an mmap'ed data file by jumping to the startOffset
        """
        assert(drv)
        self.drv = drv
        self.data = mmapData
        self.start = startOffset
        return

    def getTextTranslateValue(self,key,val):
        """ Lookup in the driver what the translation for a value should be

        e.g. 'TraceId' 

        """
        table = self.drv.traceHeaderTables[key]
        return table[val]

    def getHeader(self,fieldName):
        """ FIX: explain yourself! """
        entry  = self.drv.traceHeader[fieldName]
        start  = entry[0] - 1 + self.start
        end    = entry[1] - 1  + self.start
        length = entry[1]-entry[0] + 1
        if length > 8:
            print "get bin header... too large.  skipping ",fieldName
            return None
        
        try:
            val = struct.unpack(structIntCodes[length],self.data[start:end+1])[0]
        except KeyError:
            print "ERROR with key: ",key, "  Length not found in table",length
            assert(False)
        return val

    def printHeader(self):
        """ Dump all header info to stdout"""

        print "  FIX:  Use the lookup table at the end of the module "

        for key in self.drv.traceHeader.keys():
            val = self.getHeader(key)
            print key, " = ",val

        key = 'TraceId';    print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))
        key = 'DataUse';    print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))
        key = 'CoordUnits'; print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))
        key = 'GainType';   print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))
        key = 'SweepType';  print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))
        key = 'TimeBasis';  print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))

        key = 'TraceId';
        print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))


        try:
            key = 'OverTravel'; print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))
        except:
            print "UNKNOWN OVER TRAVEL!!!!!"

        try:
            key = 'TraceUnits'; print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))
        except:
            print "UNKNOWN TRACE UNITS!!!!!"

        try:
            key = 'TransdUnits'; print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))
        except:
            print "UNKOWN TransdUnits"
            
        key = 'SrcTypeOrient'; print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))
        key = 'SrcMeasUnits';  print key, ' := ',self.getTextTranslateValue(key,self.getHeader(key))


######################################################################
# Test code

class TestSegy(unittest.TestCase):
    """
    Run the segy class through its paces

    Use one of the BPSIO xstar segy files.  These are not proper segy
    """
    def setUp(self):
        print "skip"
        return
        #self.xstar = Segy("bpsio04L03.xstar")
        self.xstar = Segy(filename="bpsio01L00.xstar")
        self.ol = Segy(filename="9315MIG.sgy")
        return

    def tearDown(self):
        #print "tearDown: in the end of the test case"
        return

    def testIntro(self):
        print "skip"
        return
        """Test the first few bytes of the file - magic number C1 or C 1 """
        #print self.xstar.getTextHeaderLine(1)
        #print self.xstar.getTextHeaderLine(2)
        #print 'CLIENT: ',self.xstar.getTextHeaderField('CLIENT')
        #print 'AREA: ',self.xstar.getTextHeaderField('AREA')

        #f = self.xstar.getBinEntry('SampleFormat')
        #print f,"==",dataFormats[f]

        #self.xstar.printBinaryHeaders()
        
        #print "------------------------------"
        #self.ol.printBinaryHeaders()
        
        self.failUnless(self.xstar.hasR1TextFileHeader())
        self.failUnless(not self.ol.hasR1TextFileHeader())
        return

    def testSegyTrace(self):
        print "skip"
        return
        """ work on the trace class """
        print "------------------------------\n  TRACE 1 \n"

        t = SegyTraceHdr(segyHdr=self.xstar,mmapData=self.xstar.data,startOffset=3600) # Load up the 1st trace... assume no ext text block
        #t.printHeader()
        #key='TraceId'; print key,': ',self.drv.traceIdCode[t.getHeader(key)]

        # look!  If we divide by 3600, we get degrees!
        x = t.getHeader('X'); print x, x/3600.
        y = t.getHeader('Y'); print y, y/3600.
        # e.g.
        # -432364 -->  -120.101111111
        #  123862 -->    34.4061111111

        #print "------------------------------\n  TRACE 1 OL \n"
        #t = SegyTrace(self.ol.data,3600) # Load up the 1st trace... assume no ext text block
        #t.printHeader()

        return


    def testSegyTrace2(self):
        print "skip"
        return
        """ work on the trace class - load in two and make sure they are different"""
        print "------------------------------\n  TRACE 1 \n"

        segy = Segy('bpsio01L00.xstar')
        t = segy.getTraceHdr()
        print "CHECK: tw getTraceHdr() start:", t.start
        key='Shotpoint'; print key,': ',t.getHeader(key)

        # look!  If we divide by 3600, we get degrees!
        x = t.getHeader('X'); print x, x/3600.
        y = t.getHeader('Y'); print y, y/3600.

        t2 = segy.getTraceHdr(1000)
        print "CHECK: tw getTraceHdr(100) start:", t2.start
        key='Shotpoint'; print key,': ',t2.getHeader(key)

        # look!  If we divide by 3600, we get degrees!
        x2 = t2.getHeader('X'); print x, x/3600.
        y2 = t2.getHeader('Y'); print y, y/3600.

        self.failUnless ( x != x2)
        self.failUnless ( y != y2)



    def testA001Open(self):
        """ Make sure we can open a file """
        segy = Segy('bpsio01L00.xstar')

    def testA002FileBinHeader(self):
        """Be able to read a value out of the file header"""
        segy = Segy('bpsio01L00.xstar')
        hdr = segy.getHeader()
        val = hdr.getBinEntry('JobId')

    def testA003KeyValueTranslation(self):
        """ Can we get the english translation of a value?"""
        segy = Segy('bpsio01L00.xstar')
        hdr = segy.getHeader()
        key = 'SampleFormat'
        #print key,'  =>  ',hdr.getTextTranslateValue(key,1)
        self.failUnless(hdr.getTextTranslateValue(key,1),'4-byte IBM floating point')

    def testA0004TraceHdr(self):
        """ Read some header fields out of a trace """
        segy = Segy('bpsio01L00.xstar')
        t = segy.getTraceHdr(1)
        x = t.getHeader('X'); print x, x/3600.
        y = t.getHeader('Y'); print y, y/3600.

    def testA0005TraceHdr100(self):
        """ Read some header fields out of a trace """
        segy = Segy('bpsio01L00.xstar')
        t = segy.getTraceHdr(100)
        x = t.getHeader('X'); print x, x/3600.
        y = t.getHeader('Y'); print y, y/3600.

    def testA0006TraceWrite(self):
        segy = Segy('bpsio01L00.xstar')
        segy.writeTraceToFile('1.dat',traceNumber=1)
        segy.writeTraceToFile('2.dat',traceNumber=2)
        segy.writeTraceToFile('3.dat',traceNumber=3)
        segy.writeTraceToFile('100.dat',traceNumber=100)

    def testA0007TracePrintHdr(self):
        segy = Segy('bpsio01L00.xstar')
        t = segy.getTraceHdr(100)
        t.printHeader()
        
########################################
# Run the test code
if __name__ == '__main__':
    print "Testing segy"

    #import doctest, segy
    #doctest.testmod(segy)
    unittest.main()

